// **************************************************
// * Zombie PCB Code v2 */ Teensy 4.0 version Prod
// **************************************************

//  15/12/23  Disable inverter if off the throttle
//  20/12/23  Added receive of 0x554 can from Arduino to carry data TBA (eg regen setting, pedal map)

//  
#define USBSerial Serial
#define ESP_Serial Serial2

#include <FlexCAN_T4.h>

enum State : uint8_t {
    VEHICLE_OFF   = 0,
    VEHICLE_IDLE  = 1,
    VEHICLE_RUN   = 2,
    VEHICLE_ERROR = 9,
};

//control variables
bool              debugMode           = true;     
unsigned long     debugInterval       =  250;
bool              testMode            = false;     //prevents sending of 287 record for debug purposes
bool  sendToAlex          = true;
bool  dirRev              = true;    //set direction, true=ok = ccw looking at shaft which is what bmw needs
int   maxTorque           = 1999;       //2000 in normal running
int   HVMin               = 200;
int   HVMax               = 400;
int   PotMin              = 550;       //start of useful range here. Prevents open circuit being valid
int   PotMax              = 2950;      //end of useful range here, prevents short being seen as full throttle
byte  inverterFunction    = 0;        //0=disabled, 3=enabled
unsigned long   send287Interval     =  50;     //works at 50ms
unsigned long   threeSendInterval   =  75;    //100 was default

// vehicle variables
bool  throttle_waitbrake    = false;
int   vehicleState          = VEHICLE_IDLE;
bool  vehicleBrake          = false;
int   vehicleTorqueRequest  = 0;
int   globTorque            = 0;
float throttleRequest       = 0;
float vin                   = 0.0;
float startsig              = 0.0;
float BrakeVolts            = 0.0;
int throttle1raw            = 0;
int throttle2raw            = 0;
int throttle1rawOld         = 0;        
int throttle2rawOld         = 0;        


//Outlander Inverter control
int   motorAmps       = 0;
int   motorTorque     = 0;
int   motorRPM        = 0;
int   motorTempPeak   = 0;
int   motorTemp1      = 0;
int   motorTemp2      = 0;
int   inverterTemp1   = 0;
int   inverterTemp2   = 0;
int   chargerTemp1    = 0;
int   chargerTemp3    = 0;
int   chargerTemp4    = 0;
int   HVbus           = 0;
int   voltageV        = 0;
int   pedal_offset    = 0;
float AuxBattVolt     = 0;
bool  inverterCanOK   = false;
long  canRX           = 0;
long  canTX           = 0;


// pins
const int PIN_GREEN_LED =  2;         // cluster seen on PCB
const int PIN_RED_LED   =  3;         // spare on PCB
const int ADC_VIN       = A0;         // PCB power reading
const int THROTTLE1_PIN = A2;         // throttle1
const int THROTTLE2_PIN = A1;         // throttle2
const int BRAKE_PIN     = A3;         // brake
const int START_PIN     = A4;         // reads 12v ignition on
const int INV_ENABLE_PIN= A7;         // inveter enable pin.

const float ADC_CONV_FACTOR = 4096 / 3.3;                   // should be (2^12) / 3.3 = 4096

// "timers"
unsigned long lastSlowProcessMillis = 0;
unsigned long lastInverterMillis    = 0;
unsigned long lastDebug             = 0;
unsigned long lastWatchdogTimout    = 0;
unsigned long lastThreeSend         = 0;

// CAN variables
FlexCAN_T4<CAN1, RX_SIZE_256, TX_SIZE_16> myCAN1;
//CAN_message_t TxMsg;

unsigned long AlexCanTxInterval   = 250;
unsigned long AlexCanTxLast       = 0; 


void setup()
{  
    USBSerial.begin(115200);
    USBSerial.println("Here comes the Zombie..."); 
    //ESP_Serial.begin(115200);

    pinMode(INV_ENABLE_PIN, OUTPUT);            
    digitalWrite(INV_ENABLE_PIN, LOW);        

    pinMode(PIN_GREEN_LED, OUTPUT);              // PCB lights
    pinMode(PIN_RED_LED, OUTPUT);                   
    digitalWrite(PIN_GREEN_LED, LOW);
    digitalWrite(PIN_RED_LED, HIGH);
    
    // CAN
    myCAN1.begin();
    myCAN1.setBaudRate(500000);
    myCAN1.enableFIFO();
    myCAN1.enableFIFOInterrupt();
    myCAN1.onReceive(canRecieve);

    // ADC
    analogReadResolution(12);

    lastDebug             = millis();
    lastInverterMillis    = millis()+100;
    lastSlowProcessMillis = millis()+ 50;

    USBSerial.println("Setup done");
}


int readThrottle()
{
    /* read and map the throttle, dual rates, filering, etc */
    // assume 12bit ADC and 3.3v max.
    // assume 1242 = 1v
    // vehicleThrottle = 0-100%
    //
    const float throttleAlpha = 0.95;   // don't change this.
    int         throttleOut   = 0;
    int         throttle1filt = 0;
    int         throttle2filt = 0;

    // throttle reads
    throttle1raw  = analogRead(THROTTLE1_PIN);
    throttle2raw  = analogRead(THROTTLE2_PIN);
    //USBSerial.print(throttle1raw);
    //USBSerial.print(" , ");
    //USBSerial.println(throttle2raw);

    // FIR Filter (filer high freq noise)
    throttle1filt = (throttle1raw * throttleAlpha) + (throttle1rawOld * (1-throttleAlpha));
    throttle2filt = (throttle2raw * throttleAlpha) + (throttle2rawOld * (1-throttleAlpha));

    //USBSerial.println(throttle1filt);
    if(throttle1filt < PotMin or throttle1filt>PotMax)
    {
        // no throttle / 0.2v dead zone.
        throttleOut = 0.0;
    } else {
        throttleOut = map(throttle1filt,PotMin,PotMax,0,100);
        //USBSerial.print("Percent throt:");
        //USBSerial.println(throttleOut);
    }
    
    // old.
    throttle1rawOld = throttle1raw;
    throttle2rawOld = throttle2raw;
    // return
    return throttleOut;
}



void canRecieve(const CAN_message_t &msg)
{
    //USBSerial.print("got-can: ");
    //USBSerial.println(msg.id,HEX);

    // watchdog for no coms.
    lastWatchdogTimout = millis();

    switch(msg.id)
    {
        case 0x299:
            inverterCanOK = true;
            motorTempPeak = (msg.buf[0] - 40);
            inverterTemp1 = (msg.buf[1] - 40);
            inverterTemp2 = (msg.buf[4] - 40);
            canRX++;
            break;
        case 0x56D:
            break;
        case 0x29A:
            break;
        case 0x554:
             USBSerial.println("** VCU Incoming");
             break;
        case 0x566:
            break;
        case 0x732:
            inverterCanOK = true;
            motorAmps     = msg.buf[0] << 8 & msg.buf[1];
            canRX++;
            break;
        case 0x567:
            break;
        case 0x667:
            break;
        case 0x668:
            break;
        case 0x733:
            inverterCanOK = true;
            motorTemp1 = (msg.buf[0] - 40);
            motorTemp2 = (msg.buf[2] - 40);
            canRX++;
            break;

        case 0x377:
            inverterCanOK = true;
            // 12v sense from Charger
            AuxBattVolt = float(((msg.buf[0] * 256) + msg.buf[1]) * 0.01);
            //chargerTemp1 = msg.buf[4] - 40;
            //chargerTemp2 = msg.buf[5] - 40;
            //chargerTemp3 = msg.buf[6] - 40;
            //avgChargerTemp = (chargerTemp1 + chargerTemp2 + chargerTemp3) / 3;
            canRX++;
            break;
        case 0x289:
            inverterCanOK = true;
            voltageV  = msg.buf[1] & 0xFFFF;                    //damien code
            HVbus     = (msg.buf[4] * 256 + msg.buf[5]); 
            motorRPM  = (msg.buf[2] * 256 + msg.buf[3] -20000);  //damiens code
            motorTorque = ((((msg.buf[0] * 256) + msg.buf[1]) - 10000) / 10);
            canRX++;
            break;
    }
}

void loop()
{
    // main loop needs to be fast and non-blocking.
    
    
    if(debugMode==true && millis() - lastDebug > debugInterval)
    {
        if (inverterCanOK==false) {USBSerial.println("No inverter can msgs");}
        //USBSerial.print(millis());
        USBSerial.print(" Mode: ");
        printVehicleState();
        USBSerial.print(", Throttle=");
        USBSerial.print(throttleRequest);
        USBSerial.print(", Brake=");
        USBSerial.print(BrakeVolts);
        USBSerial.print("v, startsig=");
        USBSerial.print(startsig);
        USBSerial.print(", torque=");
        USBSerial.print(vehicleTorqueRequest);
        USBSerial.print(", NM send=");
        USBSerial.print(globTorque);
        USBSerial.print(", mtemp1=");
        USBSerial.print(motorTemp1);
        USBSerial.print(", mtemp2=");
        USBSerial.print(motorTemp2);
        USBSerial.print(", rpm=");
        USBSerial.print(motorRPM);
        USBSerial.print(", amps=");
        USBSerial.print(motorAmps);
        USBSerial.print(", hv=");
        USBSerial.print(HVbus);
        USBSerial.print(", vin=");
        USBSerial.print(vin);
        USBSerial.print(", volts=");
        USBSerial.print(voltageV);
        USBSerial.print(", mtorque=");
        USBSerial.print(motorTorque);

        USBSerial.println("");        

        lastDebug = millis();
    }

    if(millis() - lastWatchdogTimout > 500)
    {
        // after 1/2 second zero variables
        // as we have no CAN to inverter.
        inverterCanOK   = false;
        motorTemp1      = 0;
        motorTemp2      = 0;
        motorAmps       = 0;
        HVbus           = 0;
        motorRPM        = 0;
        AuxBattVolt     = 0;
        motorTorque     = 0;
        inverterTemp1   = 0;
        inverterTemp2   = 0;
    }

    // low speed/priority tasks
    if(millis() - lastSlowProcessMillis > 1000)
    {
        // slow stuff here.
        vin = (analogRead(ADC_VIN) / ADC_CONV_FACTOR) * 5.7;
        lastSlowProcessMillis = millis();
    }

    // inverter processes
    if(millis() - lastInverterMillis > send287Interval)
    {
          // read stuff
        startsig      = (analogRead(START_PIN) / ADC_CONV_FACTOR) * 5.7;
        BrakeVolts    = analogRead(BRAKE_PIN) / ADC_CONV_FACTOR;
        

        // process inverter messages
        if(BrakeVolts > 1.65)
        {
            // force zero throttle.
            throttle_waitbrake = true;
            throttleRequest = 0.0;
            inverterFunction=0;
        }
        else
        {
            // read throttle
            throttleRequest = readThrottle();
            inverterFunction=3;

            if(throttleRequest == 0)
            {
                // clear flag if throttle - none.
                throttle_waitbrake = false;
                inverterFunction=0;
            }
        }

        if(vehicleState == VEHICLE_RUN)
        {
            // map to torque request 
            if(throttle_waitbrake == false)
            {
                // scale the throttle to the torques
                vehicleTorqueRequest = (throttleRequest*2000)/100.0;
            }
            else
            {
                // waiting for zero throttle first after brake
                vehicleTorqueRequest = 0;
            }
        }
        sendCAN287(vehicleTorqueRequest);

        //USBSerial.println(millis()-lastInverterMillis);
        lastInverterMillis = millis();
    }
            
    if (millis()-lastThreeSend>threeSendInterval) 
    {
            combinedSend();
            lastThreeSend=millis();
    }


    // process state.
    switch(vehicleState)
    {
        case VEHICLE_OFF:
            // un-used
            break;

        case VEHICLE_IDLE:
            // here this board is live
            if (throttleRequest==0 && startsig > 3.0)
            {
                vehicleState=VEHICLE_RUN;
                digitalWrite(PIN_GREEN_LED, LOW);
                digitalWrite(PIN_RED_LED, HIGH);
                digitalWrite(INV_ENABLE_PIN, HIGH);
                inverterFunction=3;   // enable inverter
            }
            break;

        case VEHICLE_RUN:
            // assumed live.
            // wait for pedals and respond.
            // nothing to do here.
            break;

        case VEHICLE_ERROR:
            inverterFunction=0;
            // fallen into error state
            // force zero throttle
            // then
            // stopInverterControl();
            // show dash alarm.
            break;
    }

    // <>

    // send data to Alex's Teensy 4.1
    if (sendToAlex==true && millis()-AlexCanTxLast>AlexCanTxInterval)
    {
      sendAlex();
    }
}

void printVehicleState()
{
    // debug
    switch(vehicleState)
    {
        case VEHICLE_OFF:
            USBSerial.print("OFF");
            break;
        case VEHICLE_IDLE:
            USBSerial.print("IDLE");
            break;
        case VEHICLE_RUN:
            USBSerial.print("RUN");
            break;
        case VEHICLE_ERROR:
            USBSerial.print("ERROR");
            break;
        default:
            USBSerial.print("VEHICLE_????");
            break;
    }
}

//argument should be in range 0-2000
void sendCAN287(int _torqueRequest)
{

    CAN_message_t TxMsg;


    // here send torque request to inverter: range is -1000 to 2000 ?
    if(_torqueRequest > maxTorque)
    {
        _torqueRequest = maxTorque;
        USBSerial.println("** limit");
    }

    if (dirRev==true) {_torqueRequest = 0 - _torqueRequest;}  // invert the value
    _torqueRequest += 10000;                                  // center it around 10000.

    globTorque=_torqueRequest;

    uint8_t torqueLoByte = lowByte(_torqueRequest);
    uint8_t torqueHibyte = highByte(_torqueRequest);

    // inverter torque request - < 20ms.
    TxMsg.buf[0] = 0x0;               // front high
    TxMsg.buf[1] = 0x0;               // front low
    TxMsg.buf[2] = torqueHibyte;      // rear high
    TxMsg.buf[3] = torqueLoByte;      // rear low
    TxMsg.buf[4] = 0x0;               // gen torque?
    TxMsg.buf[5] = 0x0;               // gen torque?
    TxMsg.buf[6] = inverterFunction;  //0=inverter disable, 3=inverter enable
    TxMsg.buf[7] = 0x0;       
    TxMsg.id     = 0x287;
    TxMsg.len    = 8;

    if (testMode==false) {myCAN1.write(TxMsg);}
    canTX++;
}



void combinedSend()
{    
    CAN_message_t msg3;
    msg3.id = 0x371;
    msg3.len = 8;
    msg3.buf[0] = 48;
    msg3.buf[1] = 0;
    msg3.buf[2] = 0;
    msg3.buf[3] = 0;
    msg3.buf[4] = 0;
    msg3.buf[5] = 0;
    msg3.buf[6] = 0;
    msg3.buf[7] = 0;
    myCAN1.write(msg3);
    delay(1);

    CAN_message_t msg4;
    msg4.id = 0x285;
    msg4.len = 8;
    msg4.buf[0] = 0;
    msg4.buf[1] = 0;
    msg4.buf[2] = 20;
    msg4.buf[3] = 57;
    msg4.buf[4] = 143;
    msg4.buf[5] = 254;
    msg4.buf[6] = 12;
    msg4.buf[7] = 16;
    myCAN1.write(msg4);
    delay(1);

    CAN_message_t msg5;
    msg5.id = 0x286;
    msg5.len = 8;
    msg5.buf[0] = 0;
    msg5.buf[1] = 0;
    msg5.buf[2] = 0;
    msg5.buf[3] = 61;
    msg5.buf[4] = 0;
    msg5.buf[5] = 0;
    msg5.buf[6] = 33;
    msg5.buf[7] = 0;
    myCAN1.write(msg5);

    canTX=canTX+3;

}

void sendAlex() 
{
    AlexCanTxLast=millis();
    
    USBSerial.println("** Send Alex");
    CAN_message_t msg6;
    msg6.buf[0] = 0x0;               
    if (startsig  >3.0)       {msg6.buf[0]=msg6.buf[0] | 1;}
    if (BrakeVolts>3.0)       {msg6.buf[0]=msg6.buf[0] | 2;}
    if (inverterCanOK==true)  {msg6.buf[0]=msg6.buf[0] | 4;}
    msg6.buf[1] = vehicleState;
    msg6.buf[2] = highByte(int(throttleRequest));
    msg6.buf[3] = lowByte(int(throttleRequest));
    msg6.buf[4] = highByte(globTorque);
    msg6.buf[5] = lowByte(globTorque);
    //msg6.buf[6] = inverterTemp1;
    //msg6.buf[7] = inverterTemp2;
    msg6.id     = 0x555;
    msg6.len    = 8;
    myCAN1.write(msg6);
  
}
